package com.alibaba.otter.canal.deployer.monitor;

import java.io.File;
import java.io.FileFilter;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.Executors;
import java.util.concurrent.ScheduledExecutorService;
import java.util.concurrent.TimeUnit;

import org.apache.commons.lang.exception.ExceptionUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.util.Assert;
import org.springframework.util.CollectionUtils;

import com.alibaba.otter.canal.common.AbstractCanalLifeCycle;
import com.alibaba.otter.canal.common.CanalLifeCycle;
import com.alibaba.otter.canal.common.utils.NamedThreadFactory;
import com.google.common.base.Function;
import com.google.common.collect.MapMaker;

/**
 * 监听基于spring配置的instance变化
 * 
 * @author jianghang 2013-2-6 下午06:23:55
 * @version 1.0.1
 */
public class SpringInstanceConfigMonitor extends AbstractCanalLifeCycle implements InstanceConfigMonitor, CanalLifeCycle {

    private static final Logger              logger               = LoggerFactory.getLogger(SpringInstanceConfigMonitor.class);
    private String                           rootConf;
    // 扫描周期，单位秒
    private long                             scanIntervalInSecond = 5;
    private InstanceAction                   defaultAction        = null;
    private Map<String, InstanceAction>      actions              = new MapMaker().makeMap();
    private Map<String, InstanceConfigFiles> lastFiles            = new MapMaker().makeComputingMap(new Function<String, InstanceConfigFiles>() {

                                                                      public InstanceConfigFiles apply(
                                                                                                       String destination) {
                                                                          return new InstanceConfigFiles(destination);
                                                                      }
                                                                  });
    private ScheduledExecutorService         executor             = Executors.newScheduledThreadPool(
                                                                                                     1,
                                                                                                     new NamedThreadFactory(
                                                                                                                            "canal-instance-scan"));

    public void start() {
        super.start();
        Assert.notNull(rootConf, "root conf dir is null!");

        executor.scheduleWithFixedDelay(new Runnable() {

            public void run() {
                try {
                    scan();
                } catch (Throwable e) {
                    logger.error("scan failed", e);
                }
            }

        }, 0, scanIntervalInSecond, TimeUnit.SECONDS);
    }

    public void stop() {
        super.stop();
        executor.shutdownNow();
        actions.clear();
        lastFiles.clear();
    }

    public void regeister(String destination, InstanceAction action) {
        if (action != null) {
            actions.put(destination, action);
        } else {
            actions.put(destination, defaultAction);
        }
    }

    public void unRegeister(String destination) {
        actions.remove(destination);
    }

    public void setRootConf(String rootConf) {
        this.rootConf = rootConf;
    }

    private void scan() {
        File rootdir = new File(rootConf);
        if (!rootdir.exists()) {
            return;
        }

        File[] instanceDirs = rootdir.listFiles(new FileFilter() {

            public boolean accept(File pathname) {
                String filename = pathname.getName();
                return pathname.isDirectory() && !"spring".equalsIgnoreCase(filename);
            }
        });

        // 扫描目录的新增
        Set<String> currentInstanceNames = new HashSet<String>();
        for (File instanceDir : instanceDirs) {
            String destination = instanceDir.getName();
            currentInstanceNames.add(destination);
            if (!actions.containsKey(destination)) {
                notifyStart(instanceDir, destination);
            }
        }

        // 判断目录是否删除
        Set<String> deleteInstanceNames = new HashSet<String>();
        for (String destination : actions.keySet()) {
            if (!currentInstanceNames.contains(destination)) {
                deleteInstanceNames.add(destination);
            }
        }
        for (String deleteInstanceName : deleteInstanceNames) {
            notifyStop(deleteInstanceName);
        }

        // 判断目录内文件的变化
        for (File instanceDir : instanceDirs) {
            String destination = instanceDir.getName();
            File[] instanceConfigs = instanceDir.listFiles();
            InstanceConfigFiles lastFile = lastFiles.get(destination);
            boolean hasChanged = judgeFileChanged(instanceConfigs, lastFile.getInstanceFiles());
            // 通知变化
            if (hasChanged) {
                notifyReload(destination);
            }

            if (hasChanged || CollectionUtils.isEmpty(lastFile.getInstanceFiles())) {
                // 更新内容
                List<FileInfo> newFileInfo = new ArrayList<FileInfo>();
                for (File instanceConfig : instanceConfigs) {
                    newFileInfo.add(new FileInfo(instanceConfig.getName(), instanceConfig.lastModified()));
                }

                lastFile.setInstanceFiles(newFileInfo);
            }
        }
    }

    private void notifyStart(File instanceDir, String destination) {
        try {
            defaultAction.start(destination);
            actions.put(destination, defaultAction);
            logger.info("auto notify start {} successful.", destination);
        } catch (Throwable e) {
            logger.error("scan add found[{}] but start failed", destination, ExceptionUtils.getFullStackTrace(e));
        }
    }

    private void notifyStop(String destination) {
        InstanceAction action = actions.remove(destination);
        try {
            action.stop(destination);
            logger.info("auto notify stop {} successful.", destination);
        } catch (Throwable e) {
            logger.error("scan delete found[{}] but stop failed", destination, ExceptionUtils.getFullStackTrace(e));
            actions.put(destination, action);// 再重新加回去，下一次scan时再执行删除
        }
    }

    private void notifyReload(String destination) {
        InstanceAction action = actions.get(destination);
        if (action != null) {
            try {
                action.reload(destination);
                logger.info("auto notify reload {} successful.", destination);
            } catch (Throwable e) {
                logger.error("scan reload found[{}] but reload failed", destination,
                             ExceptionUtils.getFullStackTrace(e));
            }
        }
    }

    private boolean judgeFileChanged(File[] instanceConfigs, List<FileInfo> fileInfos) {
        boolean hasChanged = false;
        for (File instanceConfig : instanceConfigs) {
            for (FileInfo fileInfo : fileInfos) {
                if (instanceConfig.getName().equals(fileInfo.getName())) {
                    hasChanged |= (instanceConfig.lastModified() != fileInfo.getLastModified());
                    if (hasChanged) {
                        return hasChanged;
                    }
                }
            }
        }

        return hasChanged;
    }

    public void setDefaultAction(InstanceAction defaultAction) {
        this.defaultAction = defaultAction;
    }

    public void setScanIntervalInSecond(long scanIntervalInSecond) {
        this.scanIntervalInSecond = scanIntervalInSecond;
    }

    public static class InstanceConfigFiles {

        private String         destination;                              // instance name
        private List<FileInfo> springFile    = new ArrayList<FileInfo>(); // spring的instance xml
        private FileInfo       rootFile;                                 // canal.properties
        private List<FileInfo> instanceFiles = new ArrayList<FileInfo>(); // instance对应的配置

        public InstanceConfigFiles(String destination){
            this.destination = destination;
        }

        public String getDestination() {
            return destination;
        }

        public void setDestination(String destination) {
            this.destination = destination;
        }

        public List<FileInfo> getSpringFile() {
            return springFile;
        }

        public void setSpringFile(List<FileInfo> springFile) {
            this.springFile = springFile;
        }

        public FileInfo getRootFile() {
            return rootFile;
        }

        public void setRootFile(FileInfo rootFile) {
            this.rootFile = rootFile;
        }

        public List<FileInfo> getInstanceFiles() {
            return instanceFiles;
        }

        public void setInstanceFiles(List<FileInfo> instanceFiles) {
            this.instanceFiles = instanceFiles;
        }

    }

    public static class FileInfo {

        private String name;
        private long   lastModified = 0;

        public FileInfo(String name, long lastModified){
            this.name = name;
            this.lastModified = lastModified;
        }

        public String getName() {
            return name;
        }

        public void setName(String name) {
            this.name = name;
        }

        public long getLastModified() {
            return lastModified;
        }

        public void setLastModified(long lastModified) {
            this.lastModified = lastModified;
        }

    }

}
